#!/opt/local/bin/python


# DETAILS:
# - Author: StJohn Piano
# - Date: 2018-10-25
# - Description: This script loads a file containing dice roll results, converts these results into bytes, and calculates whether this entropy is enough to match the desired amount of entropy specified in the controls section (and if it is not, suggests how many more dice rolls will be necessary).


# ENVIRONMENTS:
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES:
# - Whitespace characters (newlines, tabs, and spaces) are permitted in the dice roll results file. 
# - Dice roll results must consist only of characters in the list "123456". 
# - Results in the list "56" are discarded. 


# INSTRUCTIONS:
# - In the CONTROLS section, set desired_n to the desired number of bytes of entropy.
# - Multiply this number by 8 to convert to bits. Divide the result by the ee_bit_rate value (1.3333) to find the expected number of dice rolls that should generate this number of entropy bits. 
# - Multiply this result by 1.1 to add a 10% margin, then round up to the nearest integer. 
# - Perform this number of dice rolls, record the results in a text file, and set the variable dice_rolls_file_path in this script to be the path to this text file. 
# - Open a terminal. Change directory to the directory containing this script. 
# - Run this script using the following command:
# 	python convert_dice_rolls_to_hex_bytes.py
# - If more dice rolls are needed in order to reach the desired amount of entropy, this script will calculate a suggested number of new dice rolls. Perform the dice rolls, add the results to the dice roll results text file, and run this script again. 



import math



def main():
	
	
	##### START CONTROLS
	dice_rolls_file_path = "dice_rolls.txt"
	desired_n = 64 # the desired amount of entropy in bytes.
	##### END CONTROLS
	
	
	##### START SETTINGS
	ee_bit_rate = 1.3333 # expected bits of entropy per dice roll
	##### END SETTINGS
	
	
	# preparation
	
	ee_byte_rate = ee_bit_rate / 8.0 # expected bytes of entropy per dice roll
	
	
	print "\n### START CONVERSION OF DICE ROLLS TO HEX BYTES\n"
	
	data = load_file(dice_rolls_file_path)
	
	if not is_printable_ascii(data):
		message = "Data contains non-printable-ascii bytes."
		stop(message)
	
	data2 = remove_whitespace(data)
	
	if not is_dice_roll_digits(data2):
		message = "Data (without whitespace) contains non-dice-roll-digit characters."
		stop(message)
		
	n2 = len(data2)
	print "- number of dice rolls: %d" % n2
		
	data3 = dice_rolls_to_base4_values(data2)
	
	n3 = len(data3)
	print "- number of dice rolls in the list [1234]: %d" % n3
	
	if n3 % 2 == 1:
		# need an even number of base4 values for conversion to base16. 
		print "- there is one extra base4 dice roll that can't be used for conversion to base16."
		data3 = data3[:-1] # remove last character from string.
	
	data4 = base4_to_base16(data3)
	
	if not is_hex(data4):
		message = "Converted data contains non-hex bytes."
		stop(message)
	
	print "- number of hex characters after conversion: %d" % len(data4)
	
	if len(data4) % 2 == 1:
		# need an even number of hex characters for conversion to entire 8-bit bytes. 
		print "- there is one extra hex character that can't be used to form an entire 8-bit byte."
		data4 = data4[:-1] # remove last character from string.
	
	n4 = len(data4) / 2
	print "- number of hex bytes: %d" % n4
	
	print "- desired number of hex bytes: %d" % desired_n
	
	if n4 >= desired_n:
		print "- the hex bytes produced are sufficient." 
		print "- hex byte output:"
		print data4
		print "- hex byte output shortened to the desired length (%d bytes):" % desired_n
		desired_bytes = data4[:desired_n*2]
		print desired_bytes
		print "- remaining hex bytes:"
		remaining = data4[desired_n*2:]
		if len(remaining) > 0:
			print remaining
		else:
			print "[None]"
		print ""
		print "Result: Desired amount of entropy (%d bytes) has been produced." % desired_n
		print "Entropy (%d bytes):" % desired_n
		print desired_bytes
		if len(remaining) > 0:
			print "Recommendation: Perhaps preserve the %d extra hex bytes in an entropy storage file." % (len(remaining) / 2)
			print "Extra hex bytes: %s" % remaining
		
	else:
		print "- the hex bytes produced are NOT sufficient."
		print "- hex byte output:"
		print data4
		m = desired_n - n4
		print "- %d hex bytes are still needed." % m
		new_rolls = float(m) / ee_byte_rate
		new_rolls = int(math.ceil(new_rolls)) # round up
		print "- expected entropy byte rate per dice roll: %.4f" % ee_byte_rate
		print "- expected necessary number of new dice rolls: %d" % new_rolls
		new_rolls2 = int(new_rolls * 1.1) + 1
		print "- expected necessary number of new dice rolls plus a 10%% margin: %d" % new_rolls2
		print ""
		print "Result: Desired amount of entropy (%d bytes) has NOT been produced." % desired_n
		print "Entropy (%d bytes):" % n4
		print data4
		print "Recommendation: Perform %d new dice rolls and add the results to the dice rolls data file, then run this script again." % new_rolls2
	
	print "\n### END CONVERSION OF DICE ROLLS TO HEX BYTES\n"
	


def base4_to_base16(data):
	if len(data) % 2 != 0:
		message = "Byte length of data is not divisible by 2."
		stop(message)
	lookup = {
		"00": "0", "01": "1", "02": "2", "03": "3",
		"10": "4", "11": "5", "12": "6", "13": "7",
		"20": "8", "21": "9", "22": "a", "23": "b",
		"30": "c", "31": "d", "32": "e", "33": "f",
		}
	output = ""
	for i in range(0, len(data), 2): # proceed two at a time.
		pair = data[i:i+2]
		base16_value = lookup[pair]
		output += base16_value
	return output


def dice_rolls_to_base4_values(data):
	# 'data' is a string containing digits representing individual dice rolls. possible digits: 123456
	output = ""
	for roll in data:
		if roll in "1234":
			# subtract 1 to move to 0-based value domain. 
			value = int(roll) - 1
			output += str(value)
		elif roll in "56":
			# unused values
			pass
	return output


def remove_whitespace(s):
	# c = character, s = string 
	whitespace = "\n\t "
	s2 = ""
	for c in s: 
		if c not in whitespace:
			s2 += c
	return s2


def is_hex(input):
	hex_characters = "0123456789abcdef"
	for c in input: # c = character
		if c not in hex_characters:
			return False
	return True


def is_dice_roll_digits(input):
	digits = "123456"
	for c in input: # c = character
		if c not in digits:
			return False
	return True


def is_printable_ascii(input):
	characters = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
	whitespace = "\n\t "
	byte_values = characters + whitespace
	for c in input: # c = character
		if c not in byte_values:
			return False
	return True


def load_file(file_path):
	import os
	if not os.path.isfile(file_path):
		message = "File not found. File path: %s" % file_path
		stop(message)
	with open(file_path) as f:
		data = f.read()
	return data


def stop(message):
	# raise an Exception to get a traceback. 
	raise Exception("\n\nERROR: %s\n" % message)


if __name__ == '__main__': main()