#!/opt/local/bin/python


# DETAILS:
# - Author: StJohn Piano
# - Date: 2018-10-05
# - Description: This script creates and signs a standard Bitcoin transaction with one input and one output.


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES:
# 
# - My working definition of a standard transaction: 
# -- It has at least one input and at least one output. 
# -- All input and output addresses are Pay-To-Public-Key-Hash (P2PKH). 
# -- All input scriptSigs contain uncompressed public keys.
#
# - All input variables for this script are strings. If any inputs are not to be used, they can be set to the empty string {""}. 
#
# - The change address is more accurately a single output sent to that address. Currently, if two transaction outputs have the change address, the transaction fee can only be subtracted from one of these outputs. 
# 
# - in input_data, value can be included as either satoshi (e.g. "satoshi_amount": "242000") or bitcoin (e.g. "bitcoin_amount": "0.00242000"). 
#
# - Hex characters in input variable values must be lowercase. 



# DEPENDENCIES
import bitcoin_functions as bf
import transaction as tr
import math




def main():
	
	
	##### START CONTROLS
	# Note: Hex characters in input variable values must be lowercase. 
	random_value = "The Face of God dhcmrlchtdj"
	# random_value must be between 1 and 32 bytes. If random_value_type is "raw_bytes", then random_value must be between 1 and 32 ASCII characters. If random_value_type is "hex_bytes", then random_value must be an even number of hex characters and between 2 and 64 hex characters (1 byte is represented by 2 hex characters). 
	# Note: Every ECDSA signature (one for each input in a transaction) requires new random entropy. 
	# random_value_type options: ["raw_bytes", "hex_bytes"]
	random_value_type = "raw_bytes"
	input_data = {
		"txid": "e4609e0f1ca854b8b07381f32ba31adbad9713205f5a4f3f56a5a32853d47855",
		"previous_output_index": "8",
		"private_key_hex": "7468655f6d6f74655f696e5f676f6427735f657965",
		#"satoshi_amount": "242000", 
		"bitcoin_amount": "0.00242", 
		}
	output_data = {
		"address": "13xPBB175FtPbPQ84iB8KuawaVy3mHrady",
		#"satoshi_amount": "200000", 
		"bitcoin_amount": "0.002", 
		}
	# note: all inputs and outputs are assumed to be Pay-To-Public-Key-Hash (P2PKH). 
	change_address = "13xPBB175FtPbPQ84iB8KuawaVy3mHrady" 
	# note: the fee will be subtracted from the amount that is being sent to the change address.
	fee = "223" # satoshi
	fee_rate = "1" # satoshi / byte
	# fee_type options: ["fee", "fee_rate"]
	fee_type = "fee_rate"
	##### END CONTROLS
	
	
	##### START SETTINGS
	
	random_value_type_options = ["raw_bytes", "hex_bytes"]
	fee_type_options = ["fee", "fee_rate"]
	
	##### END SETTINGS
	
	
	
	# preparation
	# some values will be converted to integers. 
	# - fee, input["previous_output_index"], input["satoshi_amount"], output["satoshi_amount"]
	# some values will be converted to floats:
	# - fee_rate
	
	bf.validate_string(random_value)
	bf.validate_string(random_value_type)
	bf.validate_string(change_address)
	bf.validate_string(fee)
	bf.validate_string(fee_rate)
	bf.validate_string(fee_type)
	
	bf.validate_item_in_list(fee_type, fee_type_options)
	
	# future: validate_string_is_digits() for e.g. fee prior to int() conversion. 
	if fee_type == "fee":
		fee = int(fee)
		bf.validate_positive_int(fee)
	elif fee_type == "fee_rate":
		fee_rate = float(fee_rate)
		if fee_rate < 0:
			message = "Fee rate is negative. Fee rate: %s" % str(fee_rate)
			stop(message)
	
	bf.validate_item_in_list(random_value_type, random_value_type_options)
	
	bf.validate_string(input_data["previous_output_index"])
	input_data["previous_output_index"] = int(input_data["previous_output_index"])
	
	if "bitcoin_amount" in input_data:
		bf.validate_bitcoin_amount(input_data["bitcoin_amount"])
		b = bf.bitcoin_to_satoshi(input_data["bitcoin_amount"])
		input_data["satoshi_amount"] = b
	elif "satoshi_amount" in input_data:
		s = input_data["satoshi_amount"]
		bf.validate_string(s)
		input_data["satoshi_amount"] = int(s)
		s = input_data["satoshi_amount"]
		input_data["bitcoin_amount"] = bf.satoshi_to_bitcoin(s)
	
	if "bitcoin_amount" in output_data:
		bf.validate_bitcoin_amount(output_data["bitcoin_amount"])
		b = bf.bitcoin_to_satoshi(output_data["bitcoin_amount"])
		output_data["satoshi_amount"] = b
	elif "satoshi_amount" in output_data:
		s = output_data["satoshi_amount"]
		bf.validate_string(s)
		output_data["satoshi_amount"] = int(s)
		s = output_data["satoshi_amount"]
		output_data["bitcoin_amount"] = bf.satoshi_to_bitcoin(s)
	
	bf.validate_list_contains_items(input_data.keys(), items=["txid", "previous_output_index", "private_key_hex", "satoshi_amount", "bitcoin_amount"])
	
	bf.validate_list_contains_items(output_data.keys(), items=["address", "satoshi_amount", "bitcoin_amount"])
	
	# future: validate_txid()
	bf.validate_string(input_data["txid"])
	bf.validate_positive_int(input_data["previous_output_index"])
	bf.validate_hex(input_data["private_key_hex"])
	bf.validate_positive_int(input_data["satoshi_amount"])
	bf.validate_bitcoin_amount(input_data["bitcoin_amount"])
	
	# future: validate_bitcoin_address()
	
	bf.validate_string(output_data["address"])
	bf.validate_positive_int(output_data["satoshi_amount"])
	bf.validate_bitcoin_amount(output_data["bitcoin_amount"])
	
	
	
	print "\n### START CREATION OF BITCOIN TRANSACTION\n"
	
	
	print "- Fee type: %s" % fee_type
	
	if fee_type == "fee":
		print "- Fee: %s (satoshi)" % fee
	elif fee_type == "fee_rate":
		print "- Fee rate: %s (satoshi / byte)" % fee_rate
		
	
	# create inputs
	
	input = tr.Input.create_from_txid(
		txid = input_data["txid"],
		previous_output_index = input_data["previous_output_index"],
		private_key_hex = input_data["private_key_hex"],
		bitcoin_amount = input_data["bitcoin_amount"],
		)
		
	
	# create outputs
	
	output = tr.Output.create_from_address(
		address = output_data["address"],
		bitcoin_amount = output_data["bitcoin_amount"],
		)
	
	
		
	
	
	inputs = [input]
	outputs = [output]
	random_values = [random_value]
	
	# if random_value type is "raw_bytes", convert to hex.
	if random_value_type == "raw_bytes":
		random_values = [bf.bytes_to_hex(value) for value in random_values]
	
	print "- Number of inputs (i.e. as-yet-unspent outputs): %d" % len(inputs)
	print "- Number of outputs: %d" % len(outputs)
	
	
	
	
	print "- Change address: %s" % change_address
	# find the change_output corresponding to the change_address.
	# future: how to handle the case where multiple transaction outputs have the change_address?
	change_output = None
	addresses = []
	for output in outputs:
		addresses.append(output.address)
		if output.address == change_address:
			change_output = output
			break
	if change_output == None:
		message = "Change address not found in output addresses."
		message += "\nOutput addresses:"
		for address in addresses:
			message += "\n- %s" % address
		message += "\nChange address: %s" % change_address
		stop(message)
	
	# Display the value to be sent to the change address. 
	print "- Amount to be sent to the change address: %s" % change_output.bitcoin_amount
	
	
	# print out input addresses and the total value being spent from each one. 
	# future: create and test a transaction that spends two inputs from the same address.
	input_addresses = {}
	for input in inputs:
		value = input.satoshi_amount
		address = input.address
		if address not in input_addresses:
			input_addresses[address] = value
		else:
			input_addresses[address] += value
	print "- Input addresses, with total-value-to-be-sent:"
	for address in input_addresses:
		satoshi = input_addresses[address]
		bitcoin = bf.satoshi_to_bitcoin(satoshi)
		print "-- %s: %s" % (address, bitcoin)
		
	
	# print out output addresses and the total value being sent to each one. 
	# future: create and test a transaction that sends two outputs to the same address.
	output_addresses = {}
	for output in outputs:
		value = output.satoshi_amount
		address = output.address
		if address not in output_addresses:
			output_addresses[address] = value
		else:
			output_addresses[address] += value
	print "- Output addresses, with total-value-to-be-received:"
	for address in output_addresses:
		satoshi = output_addresses[address]
		bitcoin = bf.satoshi_to_bitcoin(satoshi)
		print "-- %s: %s" % (address, bitcoin)
	
	
	# confirm that total input value >= total output value
	input_total = 0 # in satoshis
	output_total = 0 # in satoshis
	for input in inputs:
		input_total += input.satoshi_amount
	for output in outputs:
		output_total += output.satoshi_amount
	print "- Total value of all inputs: %s" % bf.satoshi_to_bitcoin(input_total)
	print "- Total value of all outputs: %s" % bf.satoshi_to_bitcoin(output_total)
	
	
	# send any extra input value (i.e. not assigned to be sent to an output) to the change_output. 
	# these amounts are in satoshis. 
	if input_total == output_total:
		print "- Total value of all inputs exactly matches total value of all outputs."
	elif input_total > output_total:
		extra = input_total - output_total
		new_amount = change_output.satoshi_amount + extra
		change_output.set_satoshi_amount(new_amount)
		print "- Total value of all inputs is greater than total value of all outputs."
		print "-- Some input value has not been assigned to an output."
		print "-- Extra value: %s" % bf.satoshi_to_bitcoin(extra)
		print "-- This extra value will be sent to the change address."
		print "-- Change address: %s" % change_output.address
		print "-- New amount to be sent to change address: %s" % change_output.bitcoin_amount
	elif input_total < output_total:
		difference = output_total - input_total
		message = "Total value of all inputs is less than total value of all outputs."
		message += "\n- Difference: %s" % bf.satoshi_to_bitcoin(lack)
		stop(message)
		
	
	# estimate size of transaction
	# we're only handling 1 input and 1 output at the moment
	input_count_size = bf.hex_len(bf.int_to_var_int(len(inputs)))
	inputs_size = len(inputs) * ( \
		32 + # previous_output_hash (32 bytes)
		4 + # previous_output_index (4 bytes)
		1 + # script_length (1-byte var_int)
		138 + # byte length of scriptSig [approximately]
		4 # sequence (4 bytes)
		)
	output_count_size = bf.hex_len(bf.int_to_var_int(len(outputs)))
	outputs_size = len(outputs) * ( \
		8 + # value
		1 + # script_length (1-byte var_int)
		25 # byte length of a standard scriptPubKey
		)
	estimated_tran_size = (
		4 + # version (4 bytes)
		input_count_size + # input_count
		inputs_size + # byte length of concatenated inputs
		output_count_size + # output_count
		outputs_size + # byte length of concatenated outputs
		4 # block lock time (4 bytes)
		)
		
	print "- Estimated transaction size: %d bytes" % estimated_tran_size
	
	
	
	# calculate approximate fee
	final_fee = None
	if fee_type == "fee":
		final_fee = fee
		print "- Final fee: %d (satoshi)" % final_fee
	elif fee_type == "fee_rate":
		final_fee = estimated_tran_size * fee_rate
		# round up to the nearest satoshi. 
		final_fee = int(math.ceil(float(final_fee)))
		print "- Fee rate: %s (satoshi / byte)" % str(fee_rate)
		print "- Calculate %d * %s and round up to nearest satoshi." % (estimated_tran_size, str(fee_rate))
		print "- Final fee: %d (satoshi)" % final_fee
		
	if final_fee == None:
		message = "Final fee not calculated."
		stop(message)
	
	print "- Final fee rate (using estimated transaction size): %.4f (satoshi per byte)" % (float(final_fee) / estimated_tran_size)
	
	# check whether the fee can be paid from the amount assigned to the change address. 
	if final_fee > change_output.satoshi_amount:
		message = "The fee is greater than the satoshi amount that is assigned to the change address."
		message += "\nFee: %d (satoshi)" % final_fee
		message += "\nAmount to be sent to change address: %d (satoshi)" % change_output.satoshi_amount
		message += "\nDifference: %d (satoshi)" % (final_fee - change_output.satoshi_amount)
		stop(message)
		
	# subtract fee from value assigned to change_output
	original_amount = change_output.satoshi_amount
	new_amount = change_output.satoshi_amount - final_fee
	change_output.set_satoshi_amount(new_amount)
	print "- Fee subtracted from amount to be sent to change address."
	print "- New amount to be sent to change address: %d (satoshi)" % change_output.satoshi_amount
	
	
	# create transaction
	tran = tr.Transaction.create(inputs, outputs)
	
	print ""
	print "Input 0:"
	print tran.inputs[0].get_description()
	
	print ""
	print "Output 0:"
	print tran.outputs[0].get_description()
	
	print ""
	print tran.get_description()
	
	print ""
	print tran.get_description_of_signable_form(input_index=0)	
	
	tran.sign(random_values)
	
	print ""
	print tran.get_description_of_signed_form()
	
	print ""
	print "Signed transaction:"
	print tran.get_signed_form()
	
	
	print "\n### END CREATION OF BITCOIN TRANSACTION\n"




def stop(message):
	bf.stop(message)


if __name__ == '__main__': main()